# sourcery skip: avoid-single-character-names-variables
"""Parses data from the Nice G.O. API.

This module provides a class to interact
with the Nice G.O. API. It allows the user
to authenticate, connect to the WebSocket
API, and interact with barriers.

Classes:
    NiceGOApi: A class to interact with the Nice G.O. API.
"""

from __future__ import annotations

import asyncio
import contextlib
import inspect
import json
import logging
from typing import Any, Callable, Coroutine, TypeVar

import aiohttp
import botocore
import yarl
from tenacity import (
    RetryCallState,
    before_sleep_log,
    retry,
    retry_base,
    retry_if_exception_type,
    wait_random_exponential,
)

from nice_go._aws_cognito_authenticator import AwsCognitoAuthenticator
from nice_go._barrier import Barrier, BarrierState, ConnectionState
from nice_go._const import ENDPOINTS_URL
from nice_go._exceptions import (
    ApiError,
    AuthFailedError,
    NoAuthError,
    ReconnectWebSocketError,
    WebSocketError,
)
from nice_go._util import find_unauthorized_error, get_request_template
from nice_go._ws_client import WebSocketClient

T = TypeVar("T")
Coro = Coroutine[Any, Any, T]
CoroT = TypeVar("CoroT", bound=Callable[..., Coro[Any]])

_LOGGER = logging.getLogger(__name__)


class _RetryIfReconnect(retry_base):
    """Retries only if reconnect is set to True."""

    def __call__(self, retry_state: RetryCallState) -> Any:
        """Check if the retry should be retried."""
        return retry_state.kwargs.get("reconnect", True)


class NiceGOApi:
    """A class to interact with the Nice G.O. API.

    This class provides methods to authenticate, connect to the WebSocket API,
    and interact with barriers.

    Attributes:
        id_token (str | None): The ID token.

    Functions:
        event: Decorator to add an event listener.
        authenticate_refresh: Authenticate using a refresh token.
        authenticate: Authenticate using username and password.
        connect: Connect to the WebSocket API.
        subscribe: Subscribe to a receiver.
        unsubscribe: Unsubscribe from a receiver.
        close: Close the connection.
        get_all_barriers: Get all barriers.
    """

    def __init__(self) -> None:
        """Initialize the NiceGOApi object."""
        self.id_token: str | None = None
        self._closing_task: asyncio.Task[None] | None = None
        self._device_ws: WebSocketClient | None = None
        self._endpoints: dict[str, Any] | None = None
        self._session: aiohttp.ClientSession | None = None
        self._event_tasks: set[asyncio.Task[None]] = set()
        self._events_ws: WebSocketClient | None = None
        self._device_connected: bool = False
        self._events_connected: bool = False
        self._events: dict[str, list[Callable[..., Coroutine[Any, Any, Any]]]] = {}

        self.event(self.on_device_connected)
        self.event(self.on_events_connected)

    async def on_device_connected(self) -> None:
        """Handle the device connected event."""
        self._device_connected = True
        if self._device_connected and self._events_connected:
            # Only dispatch when both feeds are connected
            self._dispatch("connected")

    async def on_events_connected(self) -> None:
        """Handle the events connected event."""
        self._events_connected = True
        if self._device_connected and self._events_connected:
            # Only dispatch when both feeds are connected
            self._dispatch("connected")

    def event(self, coro: CoroT) -> CoroT:
        """Decorator to add an event listener. Just a wrapper around `listen`.

        Info:
            This can only decorate coroutine functions.

        Args:
            coro (CoroT): The coroutine function to decorate.

        Examples:
            You can use this decorator to add event listeners to the API object.
            Some events include but are not limited to:

            - `connection_lost`: Triggered when the connection to the WebSocket API is
                lost.
            - `connected`: Triggered when the connection to the WebSocket API is
                established.
            - `data`: Triggered when data is received from an active subscription.
                See `subscribe`.

            >>> @api.event
            ... async def on_data(
            ...     data: dict[str, Any] | None = None,
            ... ) -> None:
            ...     if data is not None:
            ...         print(data)
        """
        self.listen(coro.__name__, coro)
        return coro

    def listen(self, event_name: str, coro: CoroT) -> Callable[[], None]:
        """Add an event listener.

        Args:
            event_name (str): The name of the event.
            coro (CoroT): The coroutine to run when the event is dispatched.

        Returns:
            A function to remove the event listener.

        Examples:
            You can use this method to add event listeners to the API object.
            Some events include but are not limited to:

            - `connection_lost`: Triggered when the connection to the WebSocket API is
                lost.
            - `connected`: Triggered when the connection to the WebSocket API is
                established.
            - `data`: Triggered when data is received from an active subscription.
                See `subscribe`.

            >>> def on_data(data: dict[str, Any] | None = None) -> None:
            ...     if data is not None:
            ...         print(data)
            ...
            >>> remove_listener = api.listen("data", on_data)
        """
        if not inspect.iscoroutinefunction(coro):
            msg = "The decorated function must be a coroutine"
            raise TypeError(msg)

        _LOGGER.debug("Adding event listener %s", coro.__name__)

        self._events.setdefault(event_name, []).append(coro)
        return lambda: self._events[event_name].remove(coro)

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        data: dict[str, Any] | None = None,
    ) -> None:
        """Run an event coroutine. For internal use only.

        Args:
            coro (Callable[..., Coroutine[Any, Any, Any]]): The coroutine to run.
            event_name (str): The name of the event.
            data (dict[str, Any] | None): The data to pass to the event coroutine.
        """
        kwargs = {}
        if data is not None:
            kwargs["data"] = data
        try:
            await coro(**kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            _LOGGER.exception("Error while handling event %s", event_name)

    def _schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        data: dict[str, Any] | None,
    ) -> None:
        """Schedule an event to be dispatched. For internal use only.

        Args:
            coro (Callable[..., Coroutine[Any, Any, Any]]): The coroutine to schedule.
            event_name (str): The name of the event.
            data (dict[str, Any] | None): The data to pass to the event coroutine.
        """
        wrapped = self._run_event(coro, event_name, data)
        task = asyncio.create_task(wrapped, name=f"NiceGO: {event_name}")
        self._event_tasks.add(task)  # See RUF006
        task.add_done_callback(self._event_tasks.discard)

    def _dispatch(self, event: str, data: dict[str, Any] | None = None) -> None:
        """Dispatch an event to listeners. For internal use only.

        Args:
            event (str): The name of the event.
            data (dict[str, Any] | None): The data to pass to the event coroutine.
        """
        method = f"on_{event}"

        coros = self._events.get(method, [])

        if not coros:
            _LOGGER.debug("No listeners for event %s", event)
            return

        _LOGGER.debug("Dispatching event %s", event)
        for coro in coros:
            self._schedule_event(coro, method, data)

    async def authenticate_refresh(
        self,
        refresh_token: str,
        session: aiohttp.ClientSession,
    ) -> None:
        """Authenticate using a previous obtained refresh token.

        Args:
            refresh_token (str): The refresh token.
            session (aiohttp.ClientSession): The client session to use.

        Raises:
            AuthFailedError: If the authentication fails.
            ApiError: If an API error occurs.
        """
        self._session = session
        await self.__authenticate(None, None, refresh_token)

    async def authenticate(
        self,
        user_name: str,
        password: str,
        session: aiohttp.ClientSession,
    ) -> str | None:
        """Authenticate using username and password.

        Args:
            user_name (str): The username.
            password (str): The password.
            session (aiohttp.ClientSession): The client session to use.

        Returns:
            The refresh token.

        Raises:
            AuthFailedError: If the authentication fails.
            ApiError: If an API error occurs.
        """
        self._session = session
        return await self.__authenticate(user_name, password, None)

    async def __authenticate(
        self,
        user_name: str | None,
        password: str | None,
        refresh_token: str | None,
    ) -> str | None:
        """Authenticate using username and password or refresh token.

        Args:
            user_name (str | None): The username.
            password (str | None): The password.
            refresh_token (str | None): The refresh token.

        Returns:
            The refresh token.

        Raises:
            AuthFailedError: If the authentication fails.
            ApiError: If an API error occurs.
        """
        try:
            _LOGGER.debug("Authenticating")

            if self._session is None:
                msg = "ClientSession not provided"
                raise ValueError(msg)

            # Get the endpoints
            data = await self._session.get(ENDPOINTS_URL)
            endpoints = await data.json()
            self._endpoints = endpoints["endpoints"]

            if self._endpoints is None:
                msg = "Endpoints not available"
                raise ApiError(msg)

            authenticator = AwsCognitoAuthenticator(
                self._endpoints["Config"]["Region"],
                self._endpoints["Config"]["ClientId"],
                self._endpoints["Config"]["UserPoolId"],
                self._endpoints["Config"]["IdentityPoolId"],
            )

            if user_name and password:
                token = await asyncio.to_thread(
                    authenticator.get_new_token,
                    user_name,
                    password,
                )
            elif refresh_token:
                token = await asyncio.to_thread(
                    authenticator.refresh_token,
                    refresh_token,
                )

            _LOGGER.debug("Authenticated")

            self.id_token = token.id_token
        except botocore.exceptions.ClientError as e:
            _LOGGER.exception("Exception while authenticating")
            if e.response["Error"]["Code"] == "NotAuthorizedException":
                raise AuthFailedError from e
            raise ApiError from e
        else:
            return token.refresh_token

    @property
    def closed(self) -> bool:
        """Check if the connection is closed."""
        return self._closing_task is not None

    async def _poll_device_ws(self) -> None:
        """Continuously polls the device WebSocket to maintain an active connection.
        This function will repeatedly call the poll method on the WebSocket if it is
        initialized.

        Returns:
            None
        """
        if self._device_ws is None:
            return
        while True:
            await self._device_ws.poll()

    async def _poll_events_ws(self) -> None:
        """Continuously polls the device WebSocket to maintain an active connection.
        This function will repeatedly call the poll method on the WebSocket if it is
        initialized.

        Returns:
            None
        """

        if self._events_ws is None:
            return
        while True:
            await self._events_ws.poll()

    async def _check_response_errors(self, response: dict[str, Any]) -> None:
        """Checks a GraphQL response for errors, namely for expired tokens.

        Args:
            response (dict[str, Any]): The response to check.

        Raises:
            AuthFailedError: If the ID token is expired.
            ApiError: If an API error occurs.

        Returns:
            None
        """

        if errors := response.get("errors"):
            if unauthorized := find_unauthorized_error(errors):
                raise AuthFailedError(unauthorized)
            raise ApiError(errors[0])

    @retry(
        wait=wait_random_exponential(multiplier=1, min=1, max=10),
        retry=_RetryIfReconnect()
        & retry_if_exception_type(
            (
                OSError,
                WebSocketError,
                aiohttp.ClientError,
                asyncio.TimeoutError,
                ReconnectWebSocketError,
            ),
        ),
        reraise=True,
        before_sleep=before_sleep_log(_LOGGER, logging.DEBUG),
    )
    async def connect(self, *, reconnect: bool = True) -> None:
        """Connect to the WebSocket API.

        Warning:
            No events will be dispatched until you subscribe to a receiver.
            Typically, you should pass the `organization` attribute of a barrier
            object to the `subscribe` method to start receiving data. Don't ask me
            why `organization` specifically, I don't know either.

        Danger:
            This method will block the event loop until the connection is closed.
            If you want to run this method in the background, you should run it in a
            separate thread or process.

        Args:
            reconnect (bool): Whether to reconnect if the connection is lost.

        Raises:
            NoAuthError: If the ID token is not available.
            ApiError: If an API error occurs.
            WebSocketError: If an error occurs while connecting.
        """
        try:
            if self.id_token is None:
                raise NoAuthError

            if self._endpoints is None:
                msg = "Endpoints not available"
                raise ApiError(msg)

            if self._session is None:
                msg = "ClientSession not provided"
                raise ValueError(msg)

            self._reconnect = reconnect

            device_url = self._endpoints["GraphQL"]["device"]["wss"]
            events_url = self._endpoints["GraphQL"]["events"]["wss"]

            _LOGGER.debug("Connecting to WebSocket API %s", device_url)

            self._device_ws = WebSocketClient(client_session=self._session)
            await self._device_ws.connect(
                self.id_token,
                yarl.URL(device_url),
                "device",
                self._dispatch,
                yarl.URL(self._endpoints["GraphQL"]["device"]["https"]).host,
            )
            self._events_ws = WebSocketClient(client_session=self._session)
            await self._events_ws.connect(
                self.id_token,
                yarl.URL(events_url),
                "events",
                self._dispatch,
                yarl.URL(self._endpoints["GraphQL"]["events"]["https"]).host,
            )

            _LOGGER.debug("Connected to WebSocket API")

            device_task = asyncio.create_task(self._poll_device_ws())
            events_task = asyncio.create_task(self._poll_events_ws())

            with contextlib.suppress(asyncio.CancelledError):
                done, pending = await asyncio.wait(
                    [device_task, events_task],
                    return_when=asyncio.FIRST_EXCEPTION,
                )

            with contextlib.suppress(UnboundLocalError):
                if exceptions := [
                    task.exception() for task in done if task.exception()
                ]:
                    for p in pending:
                        p.cancel()
                    # Make sure both WS are closed
                    await self._events_ws.close()
                    await self._device_ws.close()
                    raise exceptions[0]  # type: ignore[misc]
        except (
            OSError,
            WebSocketError,
            aiohttp.ClientError,
            asyncio.TimeoutError,
            ReconnectWebSocketError,
        ) as e:
            self._dispatch("connection_lost", {"exception": e})
            self._device_connected = False
            self._events_connected = False
            if not reconnect:
                _LOGGER.debug("Connection lost, not reconnecting")
                await self.close()
                raise

            if self.closed:
                return

            _LOGGER.debug("Connection lost, retrying...")

            # Raising triggers retry
            raise

    async def subscribe(self, receiver: str) -> list[str]:
        """Subscribe to a receiver.

        Args:
            receiver (str): The receiver to subscribe to.

        Returns:
            The subscription IDs. You can pass this into the `unsubscribe` method to
                unsubscribe from the receiver.

        Raises:
            WebSocketError: If no WebSocket connection is available.
        """
        if self._device_ws is None:
            msg = "No WebSocket connection"
            raise WebSocketError(msg)
        if self._events_ws is None:
            msg = "No WebSocket connection"
            raise WebSocketError(msg)

        _LOGGER.debug("Subscribing to receiver %s", receiver)

        return [
            await self._device_ws.subscribe(receiver),
            await self._events_ws.subscribe(receiver),
        ]

    async def unsubscribe(self, subscription_id: str) -> None:
        """Unsubscribe from a receiver.

        Args:
            subscription_id (str): The subscription ID to unsubscribe from.

        Raises:
            WebSocketError: If no WebSocket connection is available
        """
        if self._device_ws is None:
            msg = "No WebSocket connection"
            raise WebSocketError(msg)
        if self._events_ws is None:
            msg = "No WebSocket connection"
            raise WebSocketError(msg)

        _LOGGER.debug("Unsubscribing from subscription %s", subscription_id)

        await self._device_ws.unsubscribe(subscription_id)
        await self._events_ws.unsubscribe(subscription_id)

    async def close(self) -> None:
        """Close the connection.

        Raises:
            NoAuthError: If the ID token is not available.
        """

        async def _close() -> None:
            if self._device_ws:
                await self._device_ws.close()
            if self._events_ws:
                await self._events_ws.close()

        _LOGGER.debug("Closing connection")

        self._closing_task = asyncio.create_task(_close())
        await self._closing_task

    async def get_all_barriers(self) -> list[Barrier]:
        """Get all barriers.

        Returns:
            A list of all barriers.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Getting all barriers")
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template("get_all_barriers", None),
        )
        data = await response.json()

        _LOGGER.debug("Got all barriers")
        _LOGGER.debug("Data: %s", data)

        await self._check_response_errors(data)

        barriers = []

        for device in data["data"]["devicesListAll"]["devices"]:
            if device["state"]["connectionState"] is not None:
                connection_state = ConnectionState(
                    device["state"]["connectionState"]["connected"],
                    device["state"]["connectionState"]["updatedTimestamp"],
                )
            else:
                connection_state = None
            barrier_state = BarrierState(
                device["state"]["deviceId"],
                json.loads(device["state"]["reported"]),
                device["state"]["timestamp"],
                device["state"]["version"],
                connection_state,
            )
            barrier = Barrier(
                device["id"],
                device["type"],
                device["controlLevel"],
                device["attr"],
                barrier_state,
                self,
            )
            barriers.append(barrier)

        return barriers

    async def open_barrier(self, barrier_id: str) -> bool:
        """Open a barrier.

        Args:
            barrier_id (str): The ID of the barrier to open.

        Returns:
            Whether the barrier was opened successfully.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Opening barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template("open_barrier", {"barrier_id": barrier_id}),
        )
        data = await response.json()

        _LOGGER.debug("Opening barrier response: %s", data)
        await self._check_response_errors(data)

        result: bool = data["data"]["devicesControl"]

        return result

    async def close_barrier(self, barrier_id: str) -> bool:
        """Close a barrier.

        Args:
            barrier_id (str): The ID of the barrier to close.

        Returns:
            Whether the barrier was closed successfully."""
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Closing barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template(
                "close_barrier",
                {"barrier_id": barrier_id},
            ),
        )
        data = await response.json()

        _LOGGER.debug("Closing barrier response: %s", data)
        await self._check_response_errors(data)

        result: bool = data["data"]["devicesControl"]

        return result

    async def light_on(self, barrier_id: str) -> bool:
        """Turn the light on.

        Args:
            barrier_id (str): The ID of the barrier to turn the light on.

        Returns:
            Whether the light was turned on successfully.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Turning light on for barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template("light_on", {"barrier_id": barrier_id}),
        )
        data = await response.json()

        _LOGGER.debug("Turning light on response: %s", data)
        await self._check_response_errors(data)

        result: bool = data["data"]["devicesControl"]

        return result

    async def light_off(self, barrier_id: str) -> bool:
        """Turn the light off.

        Args:
            barrier_id (str): The ID of the barrier to turn the light off.

        Returns:
            Whether the light was turned off successfully.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Turning light off for barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template("light_off", {"barrier_id": barrier_id}),
        )
        data = await response.json()

        _LOGGER.debug("Turning light off response: %s", data)
        await self._check_response_errors(data)

        result: bool = data["data"]["devicesControl"]

        return result

    async def vacation_mode_on(self, barrier_id: str) -> None:
        """Turn vacation mode on.

        Args:
            barrier_id (str): The ID of the barrier to turn vacation mode on.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Turning vacation mode on for barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template(
                "vacation_mode_on",
                {"barrier_id": barrier_id},
            ),
        )
        await response.json()

    async def vacation_mode_off(self, barrier_id: str) -> None:
        """Turn vacation mode off.

        Args:
            barrier_id (str): The ID of the barrier to turn vacation mode off.

        Raises:
            NoAuthError: If the ID token is not available.
            ValueError: If the ClientSession or Endpoints are not available.
            ApiError: If an API error occurs.
        """
        if self.id_token is None:
            raise NoAuthError
        if self._session is None:
            msg = "ClientSession not provided"
            raise ValueError(msg)
        if self._endpoints is None:
            msg = "Endpoints not available"
            raise ApiError(msg)

        api_url = self._endpoints["GraphQL"]["device"]["https"]

        _LOGGER.debug("Turning vacation mode off for barrier %s", barrier_id)
        _LOGGER.debug("API URL: %s", api_url)

        headers = {"Authorization": self.id_token, "Content-Type": "application/json"}

        response = await self._session.post(
            api_url,
            headers=headers,
            json=await get_request_template(
                "vacation_mode_off",
                {"barrier_id": barrier_id},
            ),
        )
        await response.json()



"""Fixtures for tests."""

# ruff: noqa: SLF001

from unittest.mock import AsyncMock, MagicMock

import pytest
from tenacity import wait_none

from nice_go._aws_cognito_authenticator import AwsCognitoAuthenticator
from nice_go._ws_client import WebSocketClient
from nice_go.nice_go_api import NiceGOApi


@pytest.fixture
def mock_api() -> NiceGOApi:
    """Mocked NiceGOApi instance."""
    api = NiceGOApi()
    api._session = AsyncMock()
    api._device_ws = AsyncMock()
    api._endpoints = {
        "GraphQL": {
            "device": {
                "wss": "wss://test",
                "https": "https://test",
            },
            "events": {
                "wss": "wss://test",
                "https": "https://test",
            },
        },
    }
    api.connect.retry.wait = wait_none()  # type: ignore[attr-defined]
    return api


@pytest.fixture
def mock_ws_client() -> WebSocketClient:
    """Mocked WebSocketClient instance."""
    ws = WebSocketClient(client_session=MagicMock())
    ws.ws = AsyncMock(closed=False)
    ws._dispatch = MagicMock()
    ws.id_token = "test_token"
    ws.host = "test_host"
    return ws


@pytest.fixture
def mock_authenticator() -> AwsCognitoAuthenticator:
    """Mocked AwsCognitoAuthenticator instance."""
    return AwsCognitoAuthenticator(
        "test_region",
        "test_client_id",
        "test_pool_id",
        "test_identity_pool_id",
    )



# ruff: noqa: SLF001
from __future__ import annotations

import asyncio
import json
from contextlib import nullcontext as does_not_raise
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest
import yarl

from nice_go import WebSocketError
from nice_go._exceptions import AuthFailedError, ReconnectWebSocketError
from nice_go._ws_client import EventListener, WebSocketClient


async def test_ws_connect(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_session = MagicMock()
    mock_session.ws_connect = AsyncMock()
    mock_session.ws_connect.return_value = AsyncMock(closed=False)
    mock_session.ws_connect.return_value.receive = AsyncMock()
    mock_session.ws_connect.return_value.receive.return_value = MagicMock(
        data=json.dumps({"type": "connection_ack"}),
    )
    mock_ws_client.client_session = mock_session
    mock_ws_client._dispatch = MagicMock()
    await mock_ws_client.connect(
        "test_token",
        yarl.URL("wss://test_endpoint"),
        "events",
        MagicMock(),
        "test_host",
    )
    mock_session.ws_connect.assert_called_once()
    mock_ws_client._dispatch.assert_called_once()
    mock_ws_client.ws.receive.assert_called_once()
    mock_ws_client.ws.send_json.assert_called_once()


async def test_ws_init_unexpected_type(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps({"type": "unexpected_type"}),
    )
    with pytest.raises(WebSocketError):
        await mock_ws_client.init()


async def test_ws_init_token_expired(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps(
            {
                "type": "connection_error",
                "payload": {
                    "errors": [
                        {
                            "errorType": "UnauthorizedException",
                            "message": "Token has expired.",
                        },
                    ],
                },
            },
        ),
    )
    with pytest.raises(AuthFailedError, match="Token has expired."):
        await mock_ws_client.init()


async def test_ws_init_connection_error_other(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps(
            {
                "type": "connection_error",
                "payload": {"errors": [{"errorType": "SomeOtherError"}]},
            },
        ),
    )
    with pytest.raises(WebSocketError):
        await mock_ws_client.init()


async def test_ws_init_connection_error_empty_errors(
    mock_ws_client: WebSocketClient,
) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps({"type": "connection_error", "payload": {"errors": []}}),
    )
    with pytest.raises(WebSocketError):
        await mock_ws_client.init()


async def test_ws_init_connection_error_missing_payload(
    mock_ws_client: WebSocketClient,
) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps({"type": "connection_error"}),
    )
    with pytest.raises(WebSocketError):
        await mock_ws_client.init()


async def test_ws_init_token_expired_not_first_error(
    mock_ws_client: WebSocketClient,
) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps(
            {
                "type": "connection_error",
                "payload": {
                    "errors": [
                        {"errorType": "SomeOtherError", "message": "Other error."},
                        {
                            "errorType": "UnauthorizedException",
                            "message": "Token has expired.",
                        },
                    ],
                },
            },
        ),
    )
    with pytest.raises(AuthFailedError, match="Token has expired."):
        await mock_ws_client.init()


async def test_ws_init_timeout(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.side_effect = asyncio.TimeoutError
    with pytest.raises(WebSocketError):
        await mock_ws_client.init()


async def test_ws_send_json(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.send_json = AsyncMock()
    await mock_ws_client.send({"type": "test_type"})
    mock_ws_client.ws.send_json.assert_called_once()


async def test_ws_send_str(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.send_str = AsyncMock()  # sourcery skip: name-type-suffix
    await mock_ws_client.send("test_message")
    mock_ws_client.ws.send_str.assert_called_once()


async def test_ws_subscribe_and_close(mock_ws_client: WebSocketClient) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    # Test both subscribing and closing, as they are closely related
    mock_ws_client.ws.send_json = AsyncMock()
    mock_ws_client.ws.close = AsyncMock()
    mock_ws_client.ws.receive = AsyncMock()

    subscription_id = "test_id"
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps({"id": subscription_id, "type": "start_ack"}),
        type=aiohttp.WSMsgType.TEXT,
    )

    mock_ws_client.api_type = "events"

    with patch("nice_go._ws_client.uuid.uuid4") as mock_uuid:
        mock_uuid.return_value = subscription_id
        subscribe_task = asyncio.create_task(mock_ws_client.subscribe("test_query"))
        await asyncio.sleep(0.1)
        mock_ws_client.ws.send_json.assert_called_once()
        await mock_ws_client.poll()
        await subscribe_task
        mock_ws_client._timeout_task = MagicMock(cancel=MagicMock())
        await mock_ws_client.close()
        call_count_should_be = 2
        assert mock_ws_client.ws.send_json.call_count == call_count_should_be
        mock_ws_client.ws.close.assert_called_once()


@pytest.mark.parametrize(
    "msg_type",
    [
        aiohttp.WSMsgType.ERROR,
        aiohttp.WSMsgType.CLOSE,
        aiohttp.WSMsgType.CLOSING,
        aiohttp.WSMsgType.CLOSED,
    ],
)
async def test_ws_poll_errors(
    mock_ws_client: WebSocketClient,
    msg_type: aiohttp.WSMsgType,
) -> None:
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(type=msg_type)
    mock_ws_client._timeout_task = MagicMock()
    with pytest.raises(WebSocketError):
        await mock_ws_client.poll()


async def test_ws_received_message(mock_ws_client: WebSocketClient) -> None:
    mock_ws_client._dispatch = MagicMock()
    mock_ws_client.api_type = "device"
    await mock_ws_client.received_message(
        json.dumps({"type": "data", "payload": "test_payload"}),
    )
    mock_ws_client._dispatch.assert_called_once()

    with pytest.raises(WebSocketError):
        await mock_ws_client.received_message(json.dumps({"type": "error"}))

    with pytest.raises(WebSocketError):
        await mock_ws_client.received_message("invalid_json")

    mock_ws_client._timeout_task = MagicMock(done=MagicMock(return_value=False))
    mock_ws_client._timeout_task.cancel = MagicMock()
    mock_ws_client._watch_keepalive = MagicMock()  # type: ignore[method-assign]

    with patch("nice_go._ws_client.asyncio.create_task") as mock_create_task:
        await mock_ws_client.received_message(json.dumps({"type": "ka"}))
        mock_create_task.assert_called_once()


async def test_ws_received_message_dispatch_listener_skip_type(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._dispatch_listeners = [
        EventListener(
            predicate=AsyncMock(return_value=True),
            event="wrong_type",
            future=MagicMock(),
            result=MagicMock(),
        ),
    ]
    mock_ws_client.api_type = "device"
    await mock_ws_client.received_message(
        json.dumps({"type": "data", "payload": "test_payload"}),
    )
    assert len(mock_ws_client._dispatch_listeners) == 1


async def test_ws_received_message_dispatch_listener_cancelled(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._dispatch_listeners = [
        EventListener(
            predicate=AsyncMock(return_value=True),
            event="data",
            future=MagicMock(cancelled=MagicMock(return_value=True)),
            result=MagicMock(),
        ),
    ]
    mock_ws_client.api_type = "device"
    await mock_ws_client.received_message(
        json.dumps({"type": "data", "payload": "test_payload"}),
    )
    assert not mock_ws_client._dispatch_listeners


async def test_ws_received_message_dispatch_listener_predicate_exception(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._dispatch_listeners = [
        EventListener(
            predicate=MagicMock(side_effect=Exception),
            event="data",
            future=MagicMock(cancelled=MagicMock(return_value=False)),
            result=MagicMock(),
        ),
    ]
    mock_ws_client.api_type = "device"
    await mock_ws_client.received_message(
        json.dumps({"type": "data", "payload": "test_payload"}),
    )
    assert not mock_ws_client._dispatch_listeners


async def test_subscribe_timeout(mock_ws_client: WebSocketClient) -> None:
    with patch("nice_go._ws_client.asyncio.wait_for") as mock_wait_for:
        mock_wait_for.side_effect = asyncio.TimeoutError
    mock_ws_client.api_type = "events"
    with pytest.raises(WebSocketError):
        await mock_ws_client.subscribe("test_query")


async def test_unsubscribe(mock_ws_client: WebSocketClient) -> None:
    mock_ws_client._subscriptions = ["test_id"]
    assert mock_ws_client.ws is not None
    assert isinstance(mock_ws_client.ws, AsyncMock)
    mock_ws_client.ws.send_json = AsyncMock()
    await mock_ws_client.unsubscribe("test_id")
    mock_ws_client.ws.send_json.assert_called_once()
    assert "test_id" not in mock_ws_client._subscriptions
    assert mock_ws_client.ws.send_json.call_args[0][0] == {
        "id": "test_id",
        "type": "stop",
    }


async def test_closed_property(mock_ws_client: WebSocketClient) -> None:
    assert not mock_ws_client.closed
    mock_ws_client.ws = MagicMock()
    mock_ws_client.ws.closed = True
    assert mock_ws_client.closed


async def test_connect_no_host(mock_ws_client: WebSocketClient) -> None:
    with pytest.raises(ValueError, match="host must be provided"):
        await mock_ws_client.connect(
            "test_token",
            yarl.URL("wss://test_endpoint"),
            "events",
            MagicMock(),
            None,
        )


@pytest.mark.parametrize(
    ("method", "args"),
    [
        ("init", ()),
        ("send", ({"type": "test_type"},)),
        ("poll", ()),
    ],
)
async def test_closed_error(
    mock_ws_client: WebSocketClient,
    method: str,
    args: tuple[dict[str, str] | None],
) -> None:
    mock_ws_client.ws = None
    with pytest.raises(WebSocketError, match="WebSocket connection is closed"):
        await getattr(mock_ws_client, method)(*args)


async def test_close_already_closed(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client.ws = MagicMock()
    mock_ws_client.ws.closed = True
    await mock_ws_client.close()
    assert mock_ws_client.ws.close.call_count == 0


async def test_received_message_no_predicate(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._dispatch_listeners = [
        EventListener(
            predicate=None,
            event="data",
            future=MagicMock(cancelled=MagicMock(return_value=False)),
            result=MagicMock(),
        ),
    ]
    mock_ws_client.api_type = "device"
    await mock_ws_client.received_message(
        json.dumps({"type": "data", "payload": "test_payload"}),
    )
    assert not mock_ws_client._dispatch_listeners


async def test_received_message_type_or_payload_missing(
    mock_ws_client: WebSocketClient,
) -> None:
    with pytest.raises(WebSocketError):
        await mock_ws_client.received_message(json.dumps({"payload": "test_payload"}))


async def test_barrier_obstructed(mock_ws_client: WebSocketClient) -> None:
    mock_ws_client.api_type = "events"
    mock_ws_client._dispatch = MagicMock()
    mock_ws_client.dispatch_message(
        {
            "type": "data",
            "payload": {
                "data": {
                    "eventsFeed": {
                        "item": {
                            "eventId": "event-error-barrier-obstructed",
                        },
                    },
                },
            },
        },
    )
    assert mock_ws_client._dispatch.call_count == 1


async def test_event_not_barrier_obstructed(mock_ws_client: WebSocketClient) -> None:
    mock_ws_client.api_type = "events"
    mock_ws_client._dispatch = MagicMock()
    mock_ws_client.dispatch_message(
        {
            "type": "data",
            "payload": {
                "data": {
                    "eventsFeed": {
                        "item": {
                            "eventId": "event-user-barrier-opened",
                        },
                    },
                },
            },
        },
    )
    assert mock_ws_client._dispatch.call_count == 0


async def test_unsubscribe_nonexistent_subscription(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._subscriptions = []
    mock_ws_client.client_session = MagicMock(send_json=AsyncMock())
    await mock_ws_client.unsubscribe("test_id")

    assert mock_ws_client.client_session.send_json.call_count == 0


async def test_watch_keepalive(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client.ws = MagicMock(closed=False)
    mock_ws_client.ws.send_json = AsyncMock()
    mock_ws_client.ws.receive = AsyncMock()
    mock_ws_client.ws.receive.return_value = MagicMock(
        data=json.dumps({"type": "ka"}),
        type=aiohttp.WSMsgType.TEXT,
    )
    mock_ws_client.ws.close = AsyncMock()
    mock_ws_client._timeout = 0.1
    mock_ws_client._timeout_task = MagicMock(cancel=MagicMock())

    with pytest.raises(ReconnectWebSocketError):
        await mock_ws_client._watch_keepalive()

    mock_ws_client.ws.close.assert_called_once()


async def test_watch_keepalive_no_ws(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client.ws = None
    with pytest.raises(WebSocketError, match="WebSocket connection is closed"):
        await mock_ws_client._watch_keepalive()


async def test_reconnect_no_ws(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client.ws = None
    with pytest.raises(WebSocketError, match="WebSocket connection is closed"):
        await mock_ws_client._reconnect()


async def test_close_timeout_task_cancelled(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._timeout_task = asyncio.create_task(asyncio.sleep(0))
    mock_ws_client._timeout_task.cancel()

    assert mock_ws_client.ws is not None
    assert not mock_ws_client.ws.closed

    assert not mock_ws_client._timeout_task.done()

    await mock_ws_client.close()


async def test_poll_timeout_task_cancelled(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client._timeout_task = asyncio.create_task(asyncio.sleep(0))
    mock_ws_client._timeout_task.cancel()

    assert mock_ws_client.ws is not None
    assert not mock_ws_client.ws.closed

    assert not mock_ws_client._timeout_task.done()

    mock_ws_client.ws = MagicMock(closed=False)
    mock_ws_client.ws.receive = AsyncMock(
        return_value=MagicMock(
            type=aiohttp.WSMsgType.CLOSED,
        ),
    )

    with pytest.raises(WebSocketError):
        await mock_ws_client.poll()


async def test_poll_ws_closed_reconnecting(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_ws_client.reconnecting = True
    mock_ws_client.ws = MagicMock(
        receive=AsyncMock(
            return_value=MagicMock(
                type=aiohttp.WSMsgType.CLOSED,
            ),
        ),
        closed=False,
    )
    with does_not_raise():
        await mock_ws_client.poll()


async def test_close_timeout_task_exception(
    mock_ws_client: WebSocketClient,
) -> None:
    mock_timeout_task = MagicMock()
    mock_timeout_task.done.return_value = False
    mock_timeout_task.cancel = MagicMock()
    mock_timeout_task.__await__ = MagicMock(side_effect=Exception("Test exception"))

    mock_ws_client._timeout_task = mock_timeout_task

    with patch("nice_go._ws_client._LOGGER") as mock_logger:
        await mock_ws_client.close()

        mock_timeout_task.cancel.assert_called_once()
        mock_timeout_task.done.assert_called_once()
        mock_logger.debug.assert_any_call("Cancelling keepalive task")
        mock_logger.exception.assert_called_once_with(
            "Exception occurred while cancelling keepalive task",
        )



from unittest.mock import AsyncMock, MagicMock

import pytest

from nice_go import Barrier
from nice_go._barrier import BarrierState


async def test_open() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(open_barrier=AsyncMock(return_value=True)),
    )
    assert await barrier.open() is True


async def test_close() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(close_barrier=AsyncMock(return_value=True)),
    )
    assert await barrier.close() is True


async def test_light_on() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(light_on=AsyncMock(return_value=True)),
    )
    assert await barrier.light_on() is True


async def test_light_off() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(light_off=AsyncMock(return_value=True)),
    )
    assert await barrier.light_off() is True


async def test_get_attr() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [{"key": "key", "value": "value"}],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(),
    )
    assert await barrier.get_attr("key") == "value"


async def test_get_attr_not_found() -> None:
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [{"key": "key", "value": "value"}],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        MagicMock(),
    )
    with pytest.raises(KeyError) as exc_info:
        await barrier.get_attr("not_found")
    assert str(exc_info.value) == "'Attribute with key not_found not found.'"


async def test_vacation_mode_on() -> None:
    mock_api = MagicMock(vacation_mode_on=AsyncMock())
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        mock_api,
    )
    await barrier.vacation_mode_on()
    mock_api.vacation_mode_on.assert_called_once_with("barrier_id")


async def test_vacation_mode_off() -> None:
    mock_api = MagicMock(vacation_mode_off=AsyncMock())
    barrier = Barrier(
        "barrier_id",
        "barrier_type",
        "control_level",
        [],
        BarrierState(
            deviceId="device_id",
            reported={"key": "value"},
            timestamp="timestamp",
            version="version",
            connectionState=None,
        ),
        mock_api,
    )
    await barrier.vacation_mode_off()
    mock_api.vacation_mode_off.assert_called_once_with("barrier_id")



from nice_go._authentication_tokens import AuthenticationTokens


async def test_authentication_tokens() -> None:
    data = {"IdToken": "test_token", "RefreshToken": "refresh_token"}
    tokens = AuthenticationTokens(data)
    assert tokens.id_token == "test_token"
    assert tokens.refresh_token == "refresh_token"
    data = {"IdToken": "test_token"}
    tokens = AuthenticationTokens(data)
    assert tokens.id_token == "test_token"
    assert tokens.refresh_token is None



from unittest.mock import patch

from nice_go._aws_cognito_authenticator import AwsCognitoAuthenticator


async def test_refresh_token(mock_authenticator: AwsCognitoAuthenticator) -> None:
    with patch("nice_go._aws_cognito_authenticator.boto3.client") as mock_boto3_client:
        mock_boto3_client.return_value.initiate_auth.return_value = {
            "AuthenticationResult": {"IdToken": "test_token"},
        }
        result = mock_authenticator.refresh_token("refresh_token")

        assert result.id_token == "test_token"


async def test_get_new_token(mock_authenticator: AwsCognitoAuthenticator) -> None:
    with patch(
        "nice_go._aws_cognito_authenticator.boto3.client",
    ) as mock_boto3_client, patch(
        "nice_go._aws_cognito_authenticator.AWSSRP",
    ) as mock_awssrp:
        mock_boto3_client.return_value.initiate_auth.return_value = {
            "ChallengeParameters": {"key": "value"},
        }
        mock_boto3_client.return_value.respond_to_auth_challenge.return_value = {
            "AuthenticationResult": {"IdToken": "test_token"},
        }
        mock_awssrp.return_value.get_auth_params.return_value = {"key": "value"}
        mock_awssrp.return_value.process_challenge.return_value = {"key": "value"}
        result = mock_authenticator.get_new_token("username", "password")

        assert result.id_token == "test_token"
        assert result.refresh_token is None



# ruff: noqa: SLF001
from __future__ import annotations

import asyncio
import json
from contextlib import suppress
from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock, patch

import botocore
import pytest

from nice_go import (
    ApiError,
    AuthFailedError,
    NiceGOApi,
    NoAuthError,
    WebSocketError,
)

GET_ALL_BARRIERS_RESPONSE = {
    "data": {
        "devicesListAll": {
            "devices": [
                {
                    "state": {
                        "connectionState": {
                            "connected": True,
                            "updatedTimestamp": "1234567890",
                        },
                        "deviceId": "test_id",
                        "desired": json.dumps({"test": "value"}),
                        "reported": json.dumps({"test": "value"}),
                        "timestamp": "1234567890",
                        "version": 1,
                    },
                    "id": "test_id",
                    "type": "test_type",
                    "controlLevel": "test_control_level",
                    "attr": [{"key": "test_key", "value": "test_value"}],
                },
            ],
        },
    },
}

GET_ALL_BARRIERS_RESPONSE_NO_CONNECTION_STATE = {
    "data": {
        "devicesListAll": {
            "devices": [
                {
                    "state": {
                        "connectionState": None,
                        "deviceId": "test_id",
                        "desired": json.dumps({"test": "value"}),
                        "reported": json.dumps({"test": "value"}),
                        "timestamp": "1234567890",
                        "version": 1,
                    },
                    "id": "test_id",
                    "type": "test_type",
                    "controlLevel": "test_control_level",
                    "attr": [{"key": "test_key", "value": "test_value"}],
                },
            ],
        },
    },
}

if TYPE_CHECKING:
    from syrupy.assertion import SnapshotAssertion


async def test_schedule_event(mock_api: NiceGOApi) -> None:
    with patch("nice_go.nice_go_api.asyncio.create_task") as mock_create_task:
        coro = AsyncMock()
        mock_api._schedule_event(coro, "test_event", {"key": "value"})
        mock_create_task.assert_called_once()
        await mock_create_task.call_args[0][0]  # Await the coroutine that was scheduled


async def test_dispatch_event(mock_api: NiceGOApi) -> None:
    coro = AsyncMock()
    mock_api._events["on_test_event"] = [coro]
    mock_api._dispatch("test_event", {"key": "value"})
    await asyncio.sleep(0)  # Allow the event loop to run
    coro.assert_called_once()


async def test_authenticate(mock_api: NiceGOApi) -> None:
    with patch("nice_go.nice_go_api.asyncio.to_thread") as mock_to_thread:
        mock_to_thread.return_value = MagicMock(
            id_token="test_token",  # noqa: S106
            refresh_token="refresh_token",  # noqa: S106
        )
        assert mock_api._session is not None
        result = await mock_api.authenticate("username", "password", mock_api._session)
        assert result == "refresh_token"
        assert mock_api.id_token == "test_token"


async def test_connect_not_authenticated(mock_api: NiceGOApi) -> None:
    mock_api.id_token = None
    with pytest.raises(NoAuthError):
        await mock_api.connect()


async def test_connect_no_endpoints(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"
    mock_api._endpoints = None
    with pytest.raises(ApiError):
        await mock_api.connect()


async def test_subscribe_no_ws(mock_api: NiceGOApi) -> None:
    mock_api._device_ws = None
    with pytest.raises(WebSocketError):
        await mock_api.subscribe("receiver")
    mock_api._device_ws = AsyncMock()
    mock_api._events_ws = None
    with pytest.raises(WebSocketError):
        await mock_api.subscribe("receiver")


async def test_unsubscribe_no_ws(mock_api: NiceGOApi) -> None:
    mock_api._device_ws = None
    with pytest.raises(WebSocketError):
        await mock_api.unsubscribe("receiver")
    mock_api._device_ws = AsyncMock()
    mock_api._events_ws = None
    with pytest.raises(WebSocketError):
        await mock_api.unsubscribe("receiver")


async def test_get_all_barriers_not_authenticated(mock_api: NiceGOApi) -> None:
    mock_api.id_token = None
    with pytest.raises(NoAuthError):
        await mock_api.get_all_barriers()


@pytest.mark.parametrize(
    ("method_name", "expected_result"),
    [
        ("open_barrier", True),
        ("close_barrier", True),
        ("light_on", True),
        ("light_off", True),
        ("vacation_mode_on", None),
        ("vacation_mode_off", None),
    ],
)
async def test_barrier_operations(
    mock_api: NiceGOApi,
    method_name: str,
    expected_result: Any,
) -> None:
    with patch("nice_go.nice_go_api.get_request_template") as mock_get_request_template:
        mock_api.id_token = "test_token"
        mock_get_request_template.return_value = {"query": method_name}
        assert mock_api._session is not None
        assert isinstance(mock_api._session, AsyncMock)
        mock_api._session.post.return_value.json.return_value = {
            "data": {"devicesControl": True},
        }
        method = getattr(mock_api, method_name)
        result = await method("barrier_id")
        assert result is expected_result


async def test_event_decorator(mock_api: NiceGOApi) -> None:
    @mock_api.event
    async def on_test_event(data: dict[str, Any]) -> None:
        pass

    assert "on_test_event" in mock_api._events
    assert mock_api._events["on_test_event"][0] == on_test_event
    mock_api._dispatch("test_event", {})


async def test_sync_event_decorator(mock_api: NiceGOApi) -> None:
    with pytest.raises(TypeError):

        @mock_api.event  # type: ignore[type-var]
        def on_test_event(data: dict[str, Any]) -> None:
            # It's impossible for this to be called because
            # the error stops it from being added
            pass  # pragma: no cover


@pytest.mark.parametrize("error", [asyncio.CancelledError, Exception])
async def test_run_event_errors(mock_api: NiceGOApi, error: Exception) -> None:
    coro = AsyncMock()
    coro.side_effect = error
    mock_api._events["on_test_event"] = [coro]
    mock_api._dispatch("test_event", {})
    await asyncio.sleep(0)  # Allow the event loop to run
    coro.assert_called_once()


async def test_dispatch_no_listener(mock_api: NiceGOApi) -> None:
    mock_api._dispatch("test_event", {})
    await asyncio.sleep(0)  # Allow the event loop to run


async def test_authenticate_refresh(mock_api: NiceGOApi) -> None:
    with patch("nice_go.nice_go_api.asyncio.to_thread") as mock_to_thread:
        mock_to_thread.return_value = MagicMock(
            id_token="test_token",  # noqa: S106
        )
        mock_api.id_token = "test_token"
        assert mock_api._session is not None
        await mock_api.authenticate_refresh(
            "refresh_token",
            session=mock_api._session,
        )


async def test_authenticate_botocore_client_error(mock_api: NiceGOApi) -> None:
    with patch("nice_go.nice_go_api.asyncio.to_thread") as mock_to_thread:
        mock_to_thread.side_effect = botocore.exceptions.ClientError(
            error_response={"Error": {"Code": "TestException"}},
            operation_name="test",
        )
        assert mock_api._session is not None
        with pytest.raises(ApiError):
            await mock_api.authenticate("username", "password", mock_api._session)


async def test_authenticate_not_authorized_exception(mock_api: NiceGOApi) -> None:
    with patch("nice_go.nice_go_api.asyncio.to_thread") as mock_to_thread:
        mock_to_thread.side_effect = botocore.exceptions.ClientError(
            error_response={"Error": {"Code": "NotAuthorizedException"}},
            operation_name="test",
        )
        assert mock_api._session is not None
        with pytest.raises(AuthFailedError):
            await mock_api.authenticate("username", "password", mock_api._session)


async def test_connect_error(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"

    with patch("nice_go.nice_go_api.WebSocketClient") as mock_ws_client:
        mock_ws_client_instance = AsyncMock()
        mock_ws_client.return_value = mock_ws_client_instance
        mock_ws_client_instance.poll.side_effect = WebSocketError()
        with pytest.raises(WebSocketError):
            await mock_api.connect(reconnect=False)
        expected_call_count = 2
        assert mock_ws_client_instance.connect.call_count == expected_call_count


async def test_connect_closed(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"

    with patch("nice_go.nice_go_api.WebSocketClient") as mock_ws_client:
        mock_ws_client_instance = AsyncMock()
        mock_ws_client.return_value = mock_ws_client_instance

        async def side_effect() -> None:
            mock_api._closing_task = asyncio.create_task(asyncio.sleep(0))
            await asyncio.sleep(0)
            raise WebSocketError

        mock_ws_client_instance.poll = AsyncMock(side_effect=side_effect)
        await mock_api.connect(reconnect=True)
        assert mock_ws_client_instance.connect.call_count == 2  # noqa: PLR2004
        await mock_api.close()


async def test_connect_reconnect(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"

    with patch("nice_go.nice_go_api.WebSocketClient") as mock_ws_client:
        mock_ws_client_instance = AsyncMock()
        mock_ws_client.return_value = mock_ws_client_instance
        mock_ws_client_instance.connect.side_effect = WebSocketError
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(mock_api.connect(reconnect=True), timeout=0.1)
        assert mock_ws_client_instance.connect.call_count > 4  # noqa: PLR2004


async def test_subscribe(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"
    mock_api._device_ws = AsyncMock()
    mock_api._device_ws.subscribe.return_value = "test_id"
    mock_api._events_ws = AsyncMock()
    mock_api._events_ws.subscribe.return_value = "test_id_2"
    result = await mock_api.subscribe("receiver")
    assert result == ["test_id", "test_id_2"]


async def test_unsubscribe(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"
    mock_api._device_ws = AsyncMock()
    mock_api._events_ws = AsyncMock()
    await mock_api.unsubscribe("test_id")
    mock_api._device_ws.unsubscribe.assert_called_once_with("test_id")
    mock_api._events_ws.unsubscribe.assert_called_once_with("test_id")


async def test_get_all_barriers(
    mock_api: NiceGOApi,
    snapshot: SnapshotAssertion,
) -> None:
    mock_api.id_token = "test_token"
    assert mock_api._session is not None
    assert isinstance(mock_api._session, AsyncMock)
    mock_api._session.post.return_value.json.return_value = GET_ALL_BARRIERS_RESPONSE
    result = await mock_api.get_all_barriers()
    # api is an object with an address that varies, so we exclude it from the snapshot
    # It's not what we're checking anyways
    # Remove the api property from all the barriers
    for barrier in result:
        barrier.api = None  # type: ignore[assignment]
    assert result == snapshot


async def test_get_all_barriers_no_connection_state(
    mock_api: NiceGOApi,
    snapshot: SnapshotAssertion,
) -> None:
    mock_api.id_token = "test_token"
    assert mock_api._session is not None
    assert isinstance(mock_api._session, AsyncMock)
    mock_api._session.post.return_value.json.return_value = (
        GET_ALL_BARRIERS_RESPONSE_NO_CONNECTION_STATE
    )
    result = await mock_api.get_all_barriers()
    for barrier in result:
        barrier.api = None  # type: ignore[assignment]

    assert result == snapshot


@pytest.mark.parametrize(
    ("method_name"),
    [
        ("open_barrier"),
        ("close_barrier"),
        ("light_on"),
        ("light_off"),
        ("vacation_mode_on"),
        ("vacation_mode_off"),
    ],
)
async def test_barrier_operations_no_auth(
    mock_api: NiceGOApi,
    method_name: str,
) -> None:
    mock_api.id_token = None
    method = getattr(mock_api, method_name)
    with pytest.raises(NoAuthError):
        await method("barrier_id")


@pytest.mark.parametrize(
    ("method_name", "args"),
    [
        ("authenticate", ("username", "password", None)),
        ("connect", ()),
        ("get_all_barriers", ()),
        ("open_barrier", ("barrier_id",)),
        ("close_barrier", ("barrier_id",)),
        ("light_on", ("barrier_id",)),
        ("light_off", ("barrier_id",)),
        ("vacation_mode_on", ("barrier_id",)),
        ("vacation_mode_off", ("barrier_id",)),
    ],
)
async def test_no_client_session(
    mock_api: NiceGOApi,
    method_name: str,
    args: tuple[str | None],
) -> None:
    mock_api._session = None
    mock_api.id_token = "test_token"
    method = getattr(mock_api, method_name)
    with pytest.raises(ValueError, match="ClientSession not provided"):
        await method(*args)


@pytest.mark.parametrize(
    ("method_name", "args"),
    [
        ("connect", ()),
        ("get_all_barriers", ()),
        ("open_barrier", ("barrier_id",)),
        ("close_barrier", ("barrier_id",)),
        ("light_on", ("barrier_id",)),
        ("light_off", ("barrier_id",)),
        ("vacation_mode_on", ("barrier_id",)),
        ("vacation_mode_off", ("barrier_id",)),
    ],
)
async def test_no_endpoints(
    mock_api: NiceGOApi,
    method_name: str,
    args: tuple[str | None],
) -> None:
    mock_api._endpoints = None
    mock_api.id_token = "test_token"
    method = getattr(mock_api, method_name)
    with pytest.raises(ApiError, match="Endpoints not available"):
        await method(*args)


async def test_auth_no_endpoints(
    mock_api: NiceGOApi,
) -> None:
    mock_api._session = AsyncMock(
        get=AsyncMock(
            return_value=AsyncMock(
                json=AsyncMock(
                    return_value={"endpoints": None},
                ),
            ),
        ),
    )

    with pytest.raises(ApiError, match="Endpoints not available"):
        await mock_api.authenticate("username", "password", mock_api._session)


async def test_on_device_connected(mock_api: NiceGOApi) -> None:
    mock_api._events_connected = True
    coro = AsyncMock()
    mock_api.on_connected = coro  # type: ignore[attr-defined]
    await mock_api.on_device_connected()


async def test_on_events_connected(mock_api: NiceGOApi) -> None:
    mock_api._device_connected = True
    coro = AsyncMock()
    mock_api.on_connected = coro  # type: ignore[attr-defined]
    await mock_api.on_events_connected()


async def test_poll_ws_no_ws(mock_api: NiceGOApi) -> None:
    mock_api._device_ws = None
    mock_api._events_ws = None
    await mock_api._poll_device_ws()
    await mock_api._poll_events_ws()


async def test_connect_poll_task_exception(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"

    with patch("nice_go.nice_go_api.asyncio.wait") as mock_wait, patch(
        "nice_go.nice_go_api.WebSocketClient",
    ) as mock_ws_client, patch("nice_go.nice_go_api.NiceGOApi._poll_device_ws"), patch(
        "nice_go.nice_go_api.NiceGOApi._poll_events_ws",
    ):
        mock_ws_client_instance = AsyncMock()
        mock_ws_client.return_value = mock_ws_client_instance
        cancel_mock = MagicMock()
        mock_wait.return_value = (
            (
                MagicMock(
                    exception=MagicMock(return_value=ValueError("test")),
                ),
            ),
            (MagicMock(cancel=cancel_mock),),
        )
        with pytest.raises(ValueError, match="test"):
            await mock_api.connect(reconnect=False)
        cancel_mock.assert_called_once()


async def test_graphql_error_response(mock_api: NiceGOApi) -> None:
    mock_api.id_token = "test_token"
    assert mock_api._session is not None
    assert isinstance(mock_api._session, AsyncMock)
    mock_api._session.post.return_value.json.return_value = {
        "errors": [{"errorType": "TestError", "message": "test_error"}],
    }
    with pytest.raises(ApiError, match="test_error"):
        await mock_api.get_all_barriers()

    mock_api._session.post.return_value.json.return_value = {
        "errors": [
            {"errorType": "UnauthorizedException", "message": "unauthorized_error"},
        ],
    }
    with pytest.raises(AuthFailedError, match="unauthorized_error"):
        await mock_api.get_all_barriers()
